/* flashenv.c:
 * This code provides a very basic alternative to uMon's typical
 * shell variable environment.
 * It assumes it owns one sector of flash and it is dedicated to
 * storage (and retrieval) of environment variables.
 * 
 * Very basic implementation assumes each name/val pair will fit
 * within one FLASHENV_CHUNK size block of flash.  
 *
 * A given pair is simply 
 * CRC16 | NULL_TERMINATED_NAME | NULL_TERMINATED_VALUE
 * padded to the end with 0xff (erased flash).
 *
 * The CRC16 is a two byte value that spans the NAME, VALUE and
 * padding after it is written to any given chunk.
 * The chunk size (FLASHENV_CHUNK) is assumed to fit evenly within
 * the sector the environment is used in.
 *
 */

#include "config.h"
#if INCLUDE_FLASHENV
#include "cli.h"
#include "flash.h"
#include "flashenv.h"
#include "genlib.h"
#include "ether.h"
#include "timer.h"
#if INCLUDE_SPIFLASH
#include "spif.h"
#endif

#ifndef INCLUDE_BOOTQUERY
#define INCLUDE_BOOTQUERY 1000
#endif

// Ideally, FLASHENV_CHUNK should be coordinated with
// ENV_ALLOC_SIZE in env.c...
#ifndef FLASHENV_CHUNK
#define FLASHENV_CHUNK 64
#endif

void
flashenv_list(int print, int set)
{
	ushort crc;
	char *fptr = FLASHENV_BASE;

#if INCLUDE_SPIFLASH
	static char beenhere;

	if ((set) && (!beenhere)) {
		beenhere = 1;
		spif_envinit(0);
	}
#endif

	if (print)
		printf("Flashenv: 0x%x\n",(long)FLASHENV_BASE);

	// We only need to do the CRC check in this function because
	// all other functions can assume that this will be called 
	// first.
	while(fptr <= FLASHENV_END-FLASHENV_CHUNK+1) {
		crc = xcrc16((unsigned char *)(fptr+2),FLASHENV_CHUNK-2);
		if (crc != *(ushort *)fptr) {
			if (*(ushort *)fptr != 0xffff)
				printf("Corrupt flashenv\n");
			break;
		}
		else {
			char *value;

			value = fptr+2;
			while(*value != 0) {
				value++;
				if (value >= FLASHENV_END)
					return;
			}
			if (print)
				printf(" %s: %s\n",fptr+2,value+1);
			if (set)
				setenv(fptr+2,value+1);
			fptr += FLASHENV_CHUNK;
		}
	}
}

int
flashenv_erase(void)
{
	return(flasherase(FLASHENV_SNUM));
}

char *
flashenv_get(char *name)
{
	char *fptr = FLASHENV_BASE;

	while(fptr <= FLASHENV_END-FLASHENV_CHUNK+1) {
		if (*(ushort *)fptr == 0xffff) 
			break;
		else {
			if (strcmp(fptr+2,name) == 0) {
				char *value = fptr+2;
				while(*value != 0) {
					value++;
					if (value >= FLASHENV_END)
						return(0);
				}
				return(value+1);
			}
			fptr += FLASHENV_CHUNK;
		}
	}

	// If no match, return 0
	return(0);
}

int
flashenv_del(char *name)
{
	char *fptr, *ram, *rbase;

	fptr = FLASHENV_BASE;
	ram = rbase = (char *)getAppRamStart();
	while(fptr <= FLASHENV_END-FLASHENV_CHUNK+1) {
		if (*(ushort *)fptr == 0xffff) 
			break;
		else {
			if (strcmp(fptr+2,name)) {
				memset(ram,0xff,FLASHENV_CHUNK);
				memcpy(ram,fptr,FLASHENV_CHUNK);
				ram += FLASHENV_CHUNK;
			}
			fptr += FLASHENV_CHUNK;
		}
	}
	flashenv_erase();
	return(AppFlashWrite((unsigned char *)FLASHENV_BASE,
				(unsigned char *)rbase,ram-rbase));
}

int
flashenv_set(char *name, char *value)
{
	unsigned char *fptr;
	int namelen, valuelen;
	char buf[FLASHENV_CHUNK];
	unsigned short *crc = (unsigned short *)buf;

	if (flashenv_get(name))
		flashenv_del(name);

	if (value == 0)
		return(0);

	namelen = strlen(name);
	valuelen = strlen(value);
	if ((namelen + valuelen + 4) >= FLASHENV_CHUNK) {
		printf("Flashenv name/val too big\n");
		return(-1); 
	}

	memset(buf,0xff,sizeof(buf));
	strcpy(buf+2,name);
	strcpy(buf+namelen+3,value);
	*crc = xcrc16((unsigned char *)buf+2,FLASHENV_CHUNK-2);

	fptr = (unsigned char *)FLASHENV_BASE;
	while(fptr <= (unsigned char *)(FLASHENV_END-FLASHENV_CHUNK+1)) {
		if (*(ushort *)fptr == 0xffff)  {
			int rc = AppFlashWrite((unsigned char *)fptr,
				(unsigned char *)buf,namelen+valuelen+4);
			return(rc);
		}
		else {
			fptr += FLASHENV_CHUNK;
		}
	}
	printf("Flashenv full\n");
	return(-1);
}

/* flashenv_boot():
 * This supports two different boot modes:
 * 1. Using environment variables, if the BOOT variable
 *    is set, then assume it is loaded with a command
 *    string and just invoke it.  This BOOT variable may
 *    have been set by the NOR-based flash environment or
 *    (if included) the SPI-flash based environment.
 * 2. If enabled, and if BOOT is not set, then call
 *    spif_startup() 
 *
 * Note that if both BOOT and spif_startup are enabled,
 * and active, BOOT will override spif_startup.
 */
void
flashenv_boot(void)
{
	char *spifbootcmd, *bootcmd;

	bootcmd = flashenv_get("BOOT");

#if INCLUDE_SPIFSTARTUP
	spifbootcmd = spif_startup_string();
#else
	spifbootcmd = 0;
#endif

	if ((bootcmd == 0) && (spifbootcmd == 0))
		return;

#if INCLUDE_BOOTQUERY
	{
		char *cmd;
		struct elapsed_tmr tmr;

		if (bootcmd)
			cmd = bootcmd;
		else
			cmd = spifbootcmd;

		printf("[%s] 'q' to abort...\n",cmd);
		startElapsedTimer(&tmr,INCLUDE_BOOTQUERY);
		while(!msecElapsed(&tmr)) {
			pollethernet();	
			if (gotachar()) {
				if (getchar() == 'q')
					return;
			}
		}
	}
#endif

#if INCLUDE_DHCPBOOT
	/* If the IPADD shell variable is set to either DHCP or BOOTP;
	 * then we assume that handshake is in progress, so we wait
	 * for it to assign an IP address...
	 */
	printf("Wait for netboot...\n");
	while(1) {
		char *env = getenv("IPADD");
		if (env == 0)
			break;
		if ((env[0] != 'D') && (env[0] != 'B'))
			break;
		pollethernet();
	}
#endif

	if (bootcmd)
		docommand(bootcmd,1);

#if INCLUDE_SPIFSTARTUP
	if (spifbootcmd)
		spif_startup();
#endif
}
#endif
